# -*- coding: utf-8 -*-
"""Digiflow validation API"""

import shutil

from pathlib import (
    Path
)

import pytest

from digiflow.digiflow_validate import (
    DDB_IGNORE_RULES_MVW,
    FAILED_ASSERT_ERROR,
    FAILED_ASSERT_OTHER,
    DigiflowDDBException,
    ddb_validation,
)

from .conftest import (
    TEST_RES
)

# switch off saxon if not available
SAXON_PY_ENABLED = True
try:
    import saxonche
except ModuleNotFoundError: 
    SAXON_PY_ENABLED = False


@pytest.fixture(name="share_it_monography")
def _fixture_share_it_monography(tmp_path):
    """Provide test fixture"""

    testroot = tmp_path / 'VALIDATIONTEST'
    testroot.mkdir()
    file_name = '1981185920_44046.xml'
    mets_source = Path(TEST_RES) / 'ocr' / file_name
    mets_target = Path(str(testroot), file_name)
    shutil.copy(str(mets_source), str(mets_target))
    return str(mets_target)


def test_ddb_validate_opendata_44046_defaults(share_it_monography):
    """Schematron validation with common monography (Aa)
    and default ignorance yield no problems
    """

    # act
    outcome = ddb_validation(share_it_monography)

    # assert
    assert len(outcome) == 0    # nothing invalid to see


@pytest.mark.skipif(not SAXON_PY_ENABLED, reason='no saxon binary')
def test_ddb_validate_opendata_44046_ignore_ident01(share_it_monography):
    """Suppress identifier validation message for 
    simple monography (Aa)"""

    # act
    outcome = ddb_validation(share_it_monography, ignore_rules=['identifier_01'])

    # assert
    assert 'info' not in outcome
    assert len(outcome) == 0


@pytest.mark.skipif(not SAXON_PY_ENABLED, reason='no saxon binary')
def test_ddb_validate_opendata_44046_without_ignorances(share_it_monography):
    """XSLT validation with common monography (Aa)
    complains about identifier type 'gbv'
    """

    # act
    outcome = ddb_validation(share_it_monography, ignore_rules=[])

    # assert
    assert len(outcome) == 1
    assert 'identifier_01' in outcome[FAILED_ASSERT_OTHER][0]
    assert 'type:gbv' in outcome[FAILED_ASSERT_OTHER][0]
    assert '265982944' in outcome[FAILED_ASSERT_OTHER][0]


@pytest.mark.skipif(not SAXON_PY_ENABLED, reason='no saxon binary')
def test_ddb_validate_kitodo2_legacy_monography_raw(tmp_path):
    """XSLT validation outcome for out-dated
    print export data from Kitodo2 legacy system
    
    Please note, that any other roles like info or warning
    (=> amdSec_13, amdSec_15) will be swalled
    """

    # arrange
    file_name = 'k2_mets_vd18_147638674.xml'
    mets_source = Path(TEST_RES) / file_name
    mets_target = Path(str(tmp_path), file_name)
    shutil.copy(str(mets_source), str(mets_target))

    # act
    with pytest.raises(DigiflowDDBException) as _dexc:
        ddb_validation(mets_target)

    # assert
    _exc_load = _dexc.value.args[0]
    assert len(_exc_load) == 2
    assert isinstance(_exc_load, list)
    # changed from 05/22 => 06/23
    # from location_02 => location_01
    assert 'location_01' in _exc_load[0]
    assert 'dmdSec_04' in _exc_load[1]

@pytest.mark.skipif(not SAXON_PY_ENABLED, reason='no saxon binary')
def test_ddb_validate_kitodo2_legacy_monography_curated(tmp_path):
    """DDB compliant data with curated METS which doesn't 
    contain former additional dmdSec anymore so only
    warnings concerning licence and digiprovMD remain
    """

    # arrange
    file_name = 'k2_mets_vd18_147638674_curated.xml'
    mets_source = Path(TEST_RES) / file_name
    mets_target = Path(str(tmp_path), file_name)
    shutil.copy(str(mets_source), str(mets_target))

    # act
    result = ddb_validation(mets_target)

    # assert 
    assert len(result) == 1
    assert FAILED_ASSERT_ERROR not in result
    assert len(result[FAILED_ASSERT_OTHER]) == 2 
    assert 'amdSec_13' in result[FAILED_ASSERT_OTHER][0]


@pytest.mark.skipif(not SAXON_PY_ENABLED, reason='no saxon binary')
def test_ddb_validate_kitodo2_menadoc_44080924x(tmp_path):
    """XSLT validation outcome for rather
    recent digitized object from menadoc retro-digi
    
    changes between 05/22 and 06/23
    * increase number of errors from 3 to 16(!)
    * assert 'originInfo_06' in result[DDB_ERROR][0]
    * assert 'location_02' in result[DDB_ERROR][1]
    * assert 'dmdSec_04' in result[DDB_ERROR][2]
    """

    # arrange
    file_name = 'k2_mets_mena_44080924X.xml'
    mets_source = Path(TEST_RES) / file_name
    mets_target = Path(str(tmp_path), file_name)
    shutil.copy(str(mets_source), str(mets_target))

    # act
    with pytest.raises(DigiflowDDBException) as _dexc:
        ddb_validation(mets_target)

    # assert
    _load = _dexc.value.args[0]
    assert len(_load) == 16
    assert 'originInfo_06' in _load[0]
    assert 'location_01' in _load[1] # before: location_02
    assert 'dmdSec_04' in _load[2]


@pytest.mark.skipif(not SAXON_PY_ENABLED, reason='no saxon binary')
def test_ddb_validate_kitodo2_vd18_153142537_raw(tmp_path):
    """XSLT validation outcome for VD18 c-stage
    without default customm ignore rules
    
    changed from 05/22 => 06/23
    * increased errors from 7 to 11
    """

    # arrange
    file_name = 'k2_mets_vd18_153142537.xml'
    mets_source = Path(TEST_RES) / file_name
    mets_target = Path(str(tmp_path), file_name)
    shutil.copy(str(mets_source), str(mets_target))

    # act
    with pytest.raises(DigiflowDDBException) as _dexc:
        ddb_validation(mets_target, ignore_rules=[])

    # assert 
    _load = _dexc.value.args[0]
    assert len(_load) == 11
    assert 'titleInfo_02' in _load[0]
    assert 'originInfo_06' in _load[1]
    assert 'originInfo_06' in _load[2]
    assert 'location_01' in _load[3]
    assert 'dmdSec_04' in _load[4]
    assert 'fileSec_02' in _load[5]
    assert 'structMapLogical_17' in _load[6]

@pytest.mark.skipif(not SAXON_PY_ENABLED, reason='no saxon binary')
def test_ddb_validate_kitodo2_vd18_153142537_dafault_ignorances(tmp_path):
    """Schematron validation outcome for VD18 F-Stage
    which considers default ignore rules for
    multivolumes like LOCTYPE HREF and alike or
    wants fileGroup@USE='DEFAULT' (which isn't present yet)
    """

    # arrange
    file_name = 'k2_mets_vd18_153142537.xml'
    mets_source = Path(TEST_RES) / file_name
    mets_target = Path(str(tmp_path), file_name)
    shutil.copy(str(mets_source), str(mets_target))

    # act
    with pytest.raises(DigiflowDDBException) as _dexc:
        ddb_validation(mets_target, ignore_rules=DDB_IGNORE_RULES_MVW)

    # assert 
    _load = _dexc.value.args[0]
    assert len(_load) == 4
    assert _load[0].startswith('[originInfo_06]')
    assert _load[1].startswith('[originInfo_06]')
    assert _load[2].startswith('[location_01')
    assert _load[3].startswith('[dmdSec_04]')


@pytest.mark.skipif(not SAXON_PY_ENABLED, reason='no saxon binary')
def test_ddb_validate_kitodo2_vd18_153142537_curated(tmp_path):
    """DDB compliant with curated METS which doesn't contain
    the former additional dmdSec anymore and also ignores
    originInfo_06 issues"""

    # arrange
    # arrange
    file_name = 'k2_mets_vd18_153142537.xml'
    mets_source = Path(TEST_RES) / file_name
    mets_target = Path(str(tmp_path), file_name)
    shutil.copy(str(mets_source), str(mets_target))

    # act
    _ignore_them = DDB_IGNORE_RULES_MVW + ['originInfo_06']
    with pytest.raises(DigiflowDDBException) as _dexc:
        ddb_validation(mets_target, ignore_rules=_ignore_them)

    # assert 
    _load = _dexc.value.args[0]
    assert len(_load) == 2
    assert _load[0].startswith('[location_01')
    assert _load[1].startswith('[dmdSec_04]')


@pytest.mark.skipif(not SAXON_PY_ENABLED, reason='no saxon binary')
def test_ddb_validate_kitodo2_vd18_153142340_raw(tmp_path):
    """Schematron validation outcome for VD18 C-Stage
    corresponding to k2_mets_153142537.xml as-it-is"""

    # arrange
    file_name = 'k2_mets_vd18_153142340.xml'
    mets_source = Path(TEST_RES) / file_name
    mets_target = Path(str(tmp_path), file_name)
    shutil.copy(str(mets_source), str(mets_target))

    # act
    with pytest.raises(DigiflowDDBException) as _dexc:
        ddb_validation(mets_target, ignore_rules=[])

    # assert 
    _load = _dexc.value.args[0]
    assert len(_load) == 3
    assert _load[0].startswith('[originInfo_06]')
    assert _load[1].startswith('[originInfo_06]')
    assert _load[2].startswith('[structMapLogical_17]')


@pytest.mark.skipif(not SAXON_PY_ENABLED, reason='no saxon binary')
def test_ddb_validate_kitodo2_vd18_153142340( tmp_path):
    """Schematron validation outcome for VD18 C-Stage
    corresponding to k2_mets_153142537.xml
    considering ULB default ignore rules"""

    # arrange
    file_name = 'k2_mets_vd18_153142340.xml'
    mets_source = Path(TEST_RES) / file_name
    mets_target = Path(str(tmp_path), file_name)
    shutil.copy(str(mets_source), str(mets_target))

    # act
    with pytest.raises(DigiflowDDBException) as _dexc:
        ddb_validation(mets_target, digi_type='Ac')

     # assert 
    _load = _dexc.value.args[0]
    assert len(_load) == 2
    assert _load[0].startswith('[originInfo_06]')
    assert _load[1].startswith('[originInfo_06]')


@pytest.mark.skipif(not SAXON_PY_ENABLED, reason='no saxon binary')
def test_ddb_validate_kitodo2_morbio_1748529021(tmp_path):
    """Schematron validation outcome for recent Morbio
    export considering ULB default ignore rules
    
    changed due 2023/06: dropped failure for amdSec_05 (licence)
    """

    # arrange
    file_name = 'k2_mets_morbio_1748529021.xml'
    mets_source = Path(TEST_RES) / file_name
    mets_target = Path(str(tmp_path), file_name)
    shutil.copy(str(mets_source), str(mets_target))

    # act
    with pytest.raises(DigiflowDDBException) as _dexc:
        ddb_validation(mets_target, digi_type='Ac')

    # assert 
    _load = _dexc.value.args[0]
    assert len(_load) ==  2    # 2 errors
    assert _load[0].startswith('[location_01]')
    assert _load[1].startswith('[dmdSec_04]') # suspicious additional dmdSec


def test_ddb_validate_newspaper(tmp_path):
    """Test recent DDB-Validation 06/23 for
    migrated newspaper issue missing
    * license information
    """

    # arrange
    the_name = 'vls_digital_3014754.zmets.xml'
    mets_source = Path(TEST_RES) / the_name
    mets_target = Path(str(tmp_path), the_name)
    shutil.copy(str(mets_source), str(mets_target))

     # act
    with pytest.raises(DigiflowDDBException) as _dexc:
        ddb_validation(mets_target, digi_type='issue')

    # assert
    _load = _dexc.value.args[0]
    assert len(_load) == 1
    assert '[amdSec_04]' in _load[0]


def test_ddb_validate_newspaper_02(tmp_path):
    """Test recent DDB-Validation 06/23 for
    migrated newspaper issue which is quite o.k.
    """

    # arrange
    the_name = 'zd1-opendata2-1516514412012-59265.xml'
    mets_source = Path(TEST_RES) / 'opendata2' / the_name
    mets_target = Path(str(tmp_path), the_name)
    shutil.copy(str(mets_source), str(mets_target))

     # act
    _result = ddb_validation(mets_target, digi_type='OZ')

    # assert
    assert len(_result) == 0



from .digflow_identifier import *
from .digiflow_io import *
from .digiflow_metadata import *
from .digiflow_generate import *
from .digiflow_export import *
from .digiflow_validate import (
    DDB_IGNORE_RULES_BASIC,
    DDB_IGNORE_RULES_MVW,
    DDB_IGNORE_RULES_NEWSPAPERS,
    FAILED_ASSERT_ERROR,
    FAILED_ASSERT_OTHER,
    DIGIS_MULTIVOLUME,
    DIGIS_NEWSPAPER,
    REPORT_FILE_XSLT,
    DigiflowDDBException,
    DigiflowTransformException,
    apply,
    ddb_validation,
    gather_failed_asserts,
)
from .validate import (
    LABEL_VALIDATOR_SCAN_CHANNEL,
    LABEL_VALIDATOR_SCAN_COMPRESSION,
    LABEL_VALIDATOR_SCAN_RESOLUTION,
    LABEL_VALIDATOR_SCAN_FILEDATA,
    UNSET_LABEL,
    UNSET_NUMBR,
    INVALID_UNSET_LABEL,
    Validator,
    Invalid,
    FSReadException,
    FSWriteException,
    Image,
    ImageMetadata,
    InvalidImageDataException,
    resource_can_be,
    group_can_read,
    group_can_write,
    validate_tiff,
)



"""Validate data on file-level"""

import stat

from pathlib import (
    Path,
)


PERMISSION_GROUP_READ = [stat.S_IRGRP]
PERMISSION_GROUP_READ_WRITE = [stat.S_IRGRP, stat.S_IWGRP]


class FSReadException(Exception):
    """Mark State process can't even 
    read file resource due ownership
    """


class FSWriteException(Exception):
    """Mark state process tries to write
    resource for modifications and/or
    sanitizing but isn't allowed
    to store data due ownership 
    restrictions"""


def _check_modus(res_path, fs_modi=None):
    """filesystem provides detailed
    information about access permissions
    check on bit-level where *all* requested
    file modi *must* match
    """

    if isinstance(res_path, str):
        res_path = Path(res_path)
    stat_result = res_path.stat()
    _fs_modi = stat_result.st_mode
    if fs_modi is None:
        fs_modi = PERMISSION_GROUP_READ
    return all(_fs_modi & m for m in fs_modi)


def resource_can_be(res_path, modi):
    """Check whether resource can
    be read/write/executed as required
    """

    return _check_modus(res_path, modi)


def group_can_read(res_path) -> bool:
    """Inspect whether resource can be read
    at least"""

    return resource_can_be(res_path, PERMISSION_GROUP_READ)


def group_can_write(res_path) -> bool:
    """Inspect whether resouce could be written
    in case to persist modifications"""

    return resource_can_be(res_path, PERMISSION_GROUP_READ_WRITE)



# -*- coding: utf-8 -*-
"""Digiflow validation API"""

import os
import shutil

from pathlib import (
    Path
)

import lxml.etree as ET

import pytest

from digiflow.digiflow_validate import (
    DigiflowDDBException,
    apply,
    gather_failed_asserts,
)

from .conftest import (
    TEST_RES
)

# switch off saxon if not available
SAXON_PY_ENABLED = True
try:
    import saxonche
except ModuleNotFoundError: 
    SAXON_PY_ENABLED = False


TEST_XSL = str(TEST_RES / 'xsl' / 'zeitungen_ulb.xsl')


@pytest.mark.skipif(not SAXON_PY_ENABLED, reason='no saxon binary')
def test_apply_xslt_issue_with_gathering(tmp_path):
    """Simple MWE to trigger transformation including
    schematron failures => none in this case, all fine
    """

    # arrange
    file_name = 'vls_digitale_9633116.zmets.xml'
    mets_source = Path(TEST_RES) / file_name
    mets_target = os.path.join(tmp_path, file_name)
    shutil.copy(str(mets_source), str(mets_target))

    # act
    _result = apply(mets_target, path_xslt=TEST_XSL, 
                    post_process=gather_failed_asserts)

    # assert
    assert _result == {}


@pytest.mark.skipif(not SAXON_PY_ENABLED, reason='no saxon binary')
def test_apply_xslt_issue_plain(tmp_path):
    """Trigger XSLT without any postprocessing =>
    result file written, check this out
    """

    # arrange
    file_name = 'vls_digitale_9633116.zmets.xml'
    mets_source = Path(TEST_RES) / file_name
    mets_target = os.path.join(tmp_path, file_name)
    shutil.copy(str(mets_source), str(mets_target))

    # act
    _result_file: Path = apply(mets_target, path_xslt=TEST_XSL)

    # assert
    assert _result_file.exists()
    _the_root = ET.parse(_result_file).getroot()
    _map = _the_root.nsmap
    _fired_rules = _the_root.xpath('//svrl:fired-rule/text()', namespaces=_map)
    assert len(_fired_rules) == 1
    
    assert _fired_rules[0] == "Datumsangaben passen: 1840-12-31 = 1840-12-31"


@pytest.fixture(name='corrupted_issue')
def _fixture_corrupted_issue(tmp_path):
    """Create issue with mismatching information
    between publication date and day@ORDERLABEL
    
    replacing
    <mods:dateIssued encoding="iso8601">1840-12-31</mods:dateIssued>
    with
    <mods:dateIssued encoding="iso8601">1840-12-30</mods:dateIssued>

    (Therefore publication origin and @ORDERLABEL no longer fit.)
    """

    file_name = 'vls_digitale_9633116.zmets.xml'
    _source = Path(TEST_RES) / file_name
    _target = os.path.join(tmp_path, file_name)
    shutil.copy(str(_source), str(_target))
    _needle = '<mods:dateIssued encoding="iso8601">1840-12-31</mods:dateIssued>'
    _replac = '<mods:dateIssued encoding="iso8601">1840-12-30</mods:dateIssued>'
    _lines = []
    with open(_target, 'r') as _org_file:
        _lines = _org_file.readlines()
    for i, _line in enumerate(_lines):
        if _needle in _line:
            _lines[i] = _line.replace(_needle, _replac)
    with open(_target, 'w') as _corrupted:
        _corrupted.writelines(_lines)
    yield _target


@pytest.mark.skipif(not SAXON_PY_ENABLED, reason='no saxon binary')
def test_apply_xslt_issue_corrupted_file(corrupted_issue):
    """Corrupted issue without post_processing
    """

    # act
    _result_file: Path = apply(corrupted_issue, path_xslt=TEST_XSL)

    # assert
    assert _result_file.exists()
    _the_root = ET.parse(_result_file).getroot()
    _map = _the_root.nsmap
    _failed = _the_root.xpath('//svrl:failed-assert', namespaces=_map)[0]
    assert "date_mets_to_mods" in _failed.get('id')
    assert "fatal" in _failed.get('role')
    assert "Logisches Datum passt nicht" in _failed[0].text


@pytest.mark.skipif(not SAXON_PY_ENABLED, reason='no saxon binary')
def test_apply_xslt_issue_corrupted_gathering_failures(corrupted_issue):
    """Simple MWE to trigger transformation including
    schematron failures => none in this case, all fine
    """

    # act
    with pytest.raises(DigiflowDDBException) as _exc:
        apply(corrupted_issue, path_xslt=TEST_XSL, 
              post_process=gather_failed_asserts)

    # assert
    _fail_str = '[date_mets_to_mods]  (Logisches Datum passt nicht zu Publikationsdatum: 1840-12-31 != 1840-12-30)'
    assert _fail_str == _exc.value.args[0][0]



"""API for validation on file-system-level"""

import shutil

from pathlib import (
    Path,
)

import pytest

from digiflow.validate import (
    LABEL_VALIDATOR_SCAN_CHANNEL,
    LABEL_VALIDATOR_SCAN_RESOLUTION,
    Image,
    CombinedScanValidator,
    ScanChannelValidator,
    ScanResolutionValidator,
    validate_tiff,
)

from .conftest import (
    TEST_RES,
)


def test_tiff_channel_depth_invalid(tmp_path):
    """What happens if feed a tiff
    with (at least) invalid channel size?
    """

    # arrange
    file_name = '43837_max_01.tif'
    file_source = Path(TEST_RES) / 'image' / file_name
    file_target = Path(str(tmp_path), file_name)
    shutil.copy(file_source, file_target)

    # act
    _outcomes = validate_tiff(file_target)

    # assert
    assert len(_outcomes.invalids) == 1
    assert 'channel: (16, 16, 16)' == _outcomes.invalids[0].info


def test_tiff_resolution_invalid(tmp_path):
    """What happens if feed a tiff
    with floating point resolution information?
    """

    # arrange
    file_name = '8736_max_01.tif'
    file_source = Path(TEST_RES) / 'image' / file_name
    file_target = Path(str(tmp_path), file_name)
    shutil.copy(file_source, file_target)

    # act
    _outcomes:CombinedScanValidator = validate_tiff(file_target)

    # assert
    assert len(_outcomes.invalids) == 2
    assert str(_outcomes.path_input).endswith(file_name)
    assert 'xRes: 470.55' == _outcomes.invalids[0].info
    assert 'yRes: 470.55' == _outcomes.invalids[1].info


def test_tiff_validate_by_labels(tmp_path):
    """Ensure that validation can be controlled
    by provided labels so in this case only
    channels are validate (although image still
    contains invalid resolution tags)
    """

    # arrange
    file_name = '8736_max_01.tif'
    file_source = Path(TEST_RES) / 'image' / file_name
    file_target = Path(str(tmp_path), file_name)
    shutil.copy(file_source, file_target)

    # act
    _outcomes:CombinedScanValidator = validate_tiff(file_target, [LABEL_VALIDATOR_SCAN_CHANNEL])

    # assert
    assert len(_outcomes.invalids) == 0


@pytest.fixture(name='img_resolution_invalid')
def _fixture_img_resolution_invalid(tmp_path):
    file_name = '8736_max_01.tif'
    file_source = Path(TEST_RES) / 'image' / file_name
    file_target = Path(str(tmp_path), file_name)
    shutil.copy(file_source, file_target)
    _img = Image(file_target)
    _img.read()
    yield _img


def test_tiffexifresolution_resolution_not_integral(img_resolution_invalid):
    """Ensure invalid resolution is recognized"""

    # arrange
    _tiff_exif_val = ScanResolutionValidator(img_resolution_invalid)

    # act
    _tiff_exif_val.valid()

    # assert
    assert _tiff_exif_val.label == LABEL_VALIDATOR_SCAN_RESOLUTION
    assert len(_tiff_exif_val.invalids) == 2
    assert 'xRes: 470.55' == _tiff_exif_val.invalids[0].info
    assert 'yRes: 470.55' == _tiff_exif_val.invalids[1].info
    

def test_tiffexifresolution_channels_ok(img_resolution_invalid):
    """Ensure channel data is valid this time"""

    # arrange
    _tiff_exif_val = ScanChannelValidator(img_resolution_invalid)

    # act
    _tiff_exif_val.valid()

    # assert
    assert _tiff_exif_val.label == LABEL_VALIDATOR_SCAN_CHANNEL
    assert len(_tiff_exif_val.invalids) == 0


def test_tiff_grayscale_newspaper_defaults_valid(tmp_path):
    """Ensure grayscale image properly recognized
    """

    # arrange
    file_name = '1667522809_J_0025_0512.tif'
    file_source = Path(TEST_RES) / 'image' / file_name
    file_target = Path(str(tmp_path), file_name)
    shutil.copy(file_source, file_target)

    # act
    _img_data: Image = validate_tiff(file_target).img_data

    # assert
    assert _img_data.metadata.channel == (8,)
    assert _img_data.metadata.xRes == 470
    assert _img_data.metadata.yRes == 470
    assert _img_data.metadata.resolution_unit == 2
    assert str(_img_data.url).endswith(file_name)


def test_tiff_grayscale_newspaper_valid(tmp_path):
    """Ensure grayscale image properly recognized
    """

    # arrange
    file_name = '1667522809_J_0025_0512.tif'
    file_source = Path(TEST_RES) / 'image' / file_name
    file_target = Path(str(tmp_path), file_name)
    shutil.copy(file_source, file_target)

    # act
    _img_data: Image = validate_tiff(file_target).img_data

    # assert
    assert _img_data.metadata.channel == (8,)
    assert _img_data.metadata.xRes == 470
    assert _img_data.metadata.yRes == 470
    assert _img_data.metadata.resolution_unit == 2
    assert str(_img_data.url).endswith(file_name)



"""Image / TIFF-data specific Validation API"""

import datetime
import io
import os

from dataclasses import (
    dataclass,
)
from hashlib import (
    sha512
)
from pathlib import (
    Path,
)
from typing import (
    List,
)

from PIL import (
    Image as PILImage,
    ImageCms,
)
from PIL.TiffImagePlugin import (
    IMAGEWIDTH,     # 256
    IMAGELENGTH,    # 257
    BITSPERSAMPLE,  # 258 (8,8,8)=8 Bit_depth per channel
    COMPRESSION,    # 259 1=uncompressed
    PHOTOMETRIC_INTERPRETATION,  # 262 1=BlackIsZero, 2=RGB
    SAMPLESPERPIXEL,    # 277 1=greyscale, 3=3-channel RGB
    X_RESOLUTION,       # 282
    Y_RESOLUTION,       # 283
    RESOLUTION_UNIT,    # 296 2=Inches,3=cm
    SOFTWARE,       # 305
    DATE_TIME,      # 306
    ARTIST,         # 315
    COPYRIGHT,      # 33423
    ICCPROFILE,     # 34675
    ImageFileDirectory_v2,
    TiffImageFile,
    IFDRational,
)

from .common import (
    LABEL_VALIDATOR_SCAN_COMBINED,
    LABEL_VALIDATOR_SCAN_FILEDATA,
    LABEL_VALIDATOR_SCAN_CHANNEL,
    LABEL_VALIDATOR_SCAN_COMPRESSION,
	LABEL_VALIDATOR_SCAN_RESOLUTION,
    LABEL_VALIDATOR_SCAN_PHOTOMETRICS,
    UNSET_LABEL,
    UNSET_NUMBR,
    INVALID_UNSET_LABEL,
    Validator,
    Invalid,
)
from .fsdata import (
    FSReadException,
    group_can_read,
)

# special, additional tag
MODEL = 272  # EXIF 272

# tied to actual derivans implementation
DEFAULT_REQUIRED_TIFF_METADATA = [
    IMAGEWIDTH,
    IMAGELENGTH,
    BITSPERSAMPLE,
    COMPRESSION,
    SAMPLESPERPIXEL,
    X_RESOLUTION,
    Y_RESOLUTION,
    RESOLUTION_UNIT,
    ICCPROFILE,
]

TIFF_METADATA_LABELS = {
    IMAGEWIDTH: 'width',
    IMAGELENGTH: 'height',
    BITSPERSAMPLE: 'channel',
    X_RESOLUTION: 'xRes',
    Y_RESOLUTION: 'yRes',
    RESOLUTION_UNIT: 'resolution_unit',
    ARTIST: 'artist',
    COMPRESSION: 'compression',
    COPYRIGHT: 'copyright',
    MODEL: 'model',
    SOFTWARE: 'software',
    DATE_TIME: 'created',
    SAMPLESPERPIXEL: 'samples_per_pixel',
    PHOTOMETRIC_INTERPRETATION: 'photometric_interpretation',
    ICCPROFILE: 'icc_profile'
}

# module constants
MIN_SCAN_FILESIZE = 1024    # Bytes
MIN_SCAN_RESOLUTION = 300   # like 300 DPI
RES_UNIT_DPI = 2            # EXIF Value for DPI
MAX_CHANNEL_DEPTH = 8       # 8 bits / Channel
DEFAULT_COMPRESSION = 1     # EXIF 259 1=uncompressed
PHOTOMETRICS = [1, 2]       # EXIF 262 1 = BlackIsZero, 2 = rgb
GREYSCALE_OR_RGB = [1, 3]   # EXIF 277 1 = greyscale, 3 = 3-channel RGB
ADOBE_PROFILE_NAME = 'Adobe RGB (1998)'  # Preferred ICC-Profile for RGB-TIF

# date formatting
DATETIME_MIX_FORMAT = "%Y-%m-%dT%H:%M:%S"
DATETIME_SRC_FORMAT = "%Y:%m:%d %H:%M:%S"


class InvalidImageDataException(Exception):
    """Mark invalid image data"""


@dataclass
class ImageMetadata:
    """Wrap image metadata information (i.e. EXIF)
    """

    width = 0
    height = 0
    compression = UNSET_NUMBR
    photometric_interpretation = UNSET_NUMBR
    samples_per_pixel = UNSET_NUMBR
    model = UNSET_LABEL
    xRes = UNSET_NUMBR
    yRes = UNSET_NUMBR
    resolution_unit = UNSET_NUMBR
    color_space = UNSET_NUMBR
    artist = UNSET_LABEL
    copyright = UNSET_LABEL
    software = UNSET_LABEL
    created = UNSET_LABEL  # EXIF TAG 306 DATE_TIME
    channel = UNSET_NUMBR  # EFIX TAG 258 BITSPERSAMPLE


@dataclass
class Image:
    """Store required data"""

    local_path: str
    url = UNSET_LABEL
    file_size = 0
    time_stamp = None
    profile = UNSET_LABEL
    image_checksum = UNSET_LABEL
    metadata: ImageMetadata = None
    invalids = []

    def read(self):
        """Try to read image data"""
        try:
            _ts_object = datetime.datetime.now()
            self.time_stamp = _ts_object.strftime(DATETIME_SRC_FORMAT)
            # open resource
            _pil_img: TiffImageFile = PILImage.open(self.local_path)
            _image_bytes = _pil_img.tobytes()
            self.image_checksum = sha512(_image_bytes).hexdigest()

            # read information from TIF TAG V2 section
            _meta_data: ImageMetadata = self._read(_pil_img)
            _meta_data.color_space = _pil_img.mode

            # datetime present?
            if _meta_data.created is None or _meta_data.created == UNSET_LABEL:
                ctime = os.stat(self.local_path).st_ctime
                dt_object = datetime.datetime.fromtimestamp(ctime)
                _meta_data.created = dt_object.strftime(DATETIME_SRC_FORMAT)
            self.metadata = _meta_data
            _pil_img.close()
        except Exception as exc:
            raise InvalidImageDataException(exc.args[0])

    def _read(self, img_data:TiffImageFile) -> ImageMetadata:
        """Read present TIFF metadata information
        Please note, ICC Profile requires special
        care since this is included in bytes
        """

        _image_tags: ImageFileDirectory_v2 = img_data.tag_v2
        _image_md = ImageMetadata()
        for _tag in _image_tags:
            if _tag in TIFF_METADATA_LABELS:
                _label = TIFF_METADATA_LABELS[_tag]
                if _tag != ICCPROFILE:
                    _val = _image_tags[_tag]
                    if isinstance(_val, str):
                        _val = str(_val).strip()
                    setattr(_image_md, _label, _val)
                else:
                    _profile = img_data.info.get('icc_profile')
                    if _profile:
                        _profile_data = io.BytesIO(_profile)
                        _profile_cms = ImageCms.ImageCmsProfile(_profile_data)
                        if _profile_cms:
                            _profile_name = ImageCms.getProfileDescription(_profile_cms)
                            if _profile_name:
                                _stripped = _profile_name.strip()
                            self.profile = _stripped
        return _image_md

    def __str__(self):
        _fsize = f"{self.file_size}"
        _md = self.metadata
        _mds_res = f"({_md.xRes},{_md.yRes},{_md.resolution_unit})"
        _mds = f"\t{_mds_res}\t{_md.created}\t{_md.artist}\t{_md.copyright}\t{_md.model}\t{_md.software}\t{self.profile}"
        _invalids = f"INVALID{self.invalids}" if len(
            self.invalids) > 0 else 'VALID'
        return f"{self.url}\t{self.image_checksum}\t{_fsize}\t{self.metadata.width}x{self.metadata.height}\t{_mds}\t{self.time_stamp}\t{_invalids}"


class ScanFileValidator(Validator):
    """Validate file on very basic level, i.e.
    whether it can be read, file size suspicous
    or no metadata tags for width/height present
    """

    def __init__(self, input_data: Path):
        super().__init__(LABEL_VALIDATOR_SCAN_FILEDATA, input_data)

    def valid(self) -> bool:
        """Fail faster: If file can't even be read by group,
        """

        if not group_can_read(self.input_data):
            raise FSReadException(f'No permission to read {self.input_data}!')
        _file_size = os.path.getsize(self.input_data)
        if _file_size < MIN_SCAN_FILESIZE:
            _msg = f"Filesize {_file_size} below {MIN_SCAN_FILESIZE}b for {self.input_data}!"
            self.invalids.append(Invalid(self.label, _msg))
        _input: Image = self.input_data
        _input.read()
        _imd: ImageMetadata = _input.metadata
        if not _imd.width or not _imd.height:
            self.invalids.append(Invalid(self.label, "No width/height set!"))
        return super().valid()


class ScanChannelValidator(Validator):
    """Validate EXIF metadata concerning
    maximal processable channel depth and
    number of channels
    """

    def __init__(self, input_data: Image):
        super().__init__(LABEL_VALIDATOR_SCAN_CHANNEL, input_data)

    def valid(self) -> bool:
        _input: Image = self.input_data
        _imd: ImageMetadata = _input.metadata
        _ssp = _imd.samples_per_pixel
        if _ssp == UNSET_LABEL:
            self.invalids.append(Invalid(self.label, f"{INVALID_UNSET_LABEL} {TIFF_METADATA_LABELS[SAMPLESPERPIXEL]}"))
        elif _ssp not in GREYSCALE_OR_RGB:
            self.invalids.append(Invalid(self.label, f"{TIFF_METADATA_LABELS[SAMPLESPERPIXEL]}: {_ssp}"))
        if not all(c <= MAX_CHANNEL_DEPTH for c in _imd.channel):
            self.invalids.append(Invalid(self.label, f"{TIFF_METADATA_LABELS[BITSPERSAMPLE]}: {_imd.channel}"))
        return super().valid()
    

class ScanCompressionValidator(Validator):
    """Validate EXIF metadata uncompressed set"""

    def __init__(self, input_data: Image):
        super().__init__(LABEL_VALIDATOR_SCAN_COMPRESSION, input_data)

    def valid(self) -> bool:
        """Check flag umcompressed data set"""

        _input: Image = self.input_data
        _imd: ImageMetadata = _input.metadata
        if _imd.compression != 1:
            self.invalids.append(Invalid(self.label, f"{TIFF_METADATA_LABELS[COMPRESSION]}: {_imd.compression}"))
        return super().valid()


class ScanResolutionValidator(Validator):
    """Validate EXIF metadata concerning
    values of required Resolution information
    * XResolution / YResolution: must not be rationals
    * Resolution Unit must not be different from DPI
    """

    def __init__(self, input_data: Image):
        super().__init__(LABEL_VALIDATOR_SCAN_RESOLUTION, input_data)

    def valid(self) -> bool:
        _input: Image = self.input_data
        _imd: ImageMetadata = _input.metadata
        _prefix = self.label
        _lbl_x = f"{TIFF_METADATA_LABELS[X_RESOLUTION]}"
        _lbl_y = f"{TIFF_METADATA_LABELS[Y_RESOLUTION]}"
        _lbl_unit = f"{TIFF_METADATA_LABELS[RESOLUTION_UNIT]}"
        if _imd.resolution_unit == UNSET_NUMBR:
            self.invalids.append(Invalid(_prefix, f"{INVALID_UNSET_LABEL} {_lbl_unit}"))
        if _imd.resolution_unit != RES_UNIT_DPI:
            self.invalids.append(Invalid(_prefix, f"{_lbl_unit} {_imd.resolution_unit} != {RES_UNIT_DPI}"))
        if _imd.xRes == UNSET_NUMBR:
            self.invalids.append(Invalid(_prefix, f"{INVALID_UNSET_LABEL} {_lbl_x}"))
        if _imd.xRes < MIN_SCAN_RESOLUTION:
            self.invalids.append(Invalid(_prefix, f"{_lbl_x}: {_imd.xRes}"))
        if _imd.yRes == UNSET_NUMBR:
            self.invalids.append(Invalid(_prefix, f"{INVALID_UNSET_LABEL} {_lbl_y}"))
        if _imd.yRes < MIN_SCAN_RESOLUTION:
            self.invalids.append(Invalid(_prefix, f"{_lbl_y}: {_imd.yRes}"))
        _xres: IFDRational = _imd.xRes
        if _xres.real.denominator != 1:
            self.invalids.append(Invalid(_prefix, f"{_lbl_x}: {_imd.xRes}"))
        _yres: IFDRational = _imd.yRes
        if _yres.real.denominator != 1:
            self.invalids.append(Invalid(_prefix, f"{_lbl_y}: {_imd.yRes}"))
        return super().valid()
    

class ScanPhotometricValidator(Validator):
    """Check EXIF Photometric information
    * Photometric interpretation must be Grayscale or RGB 
    * if RGB, then profile must be ADOBE RGB 1998
    """

    def __init__(self, input_data: Image):
        super().__init__(LABEL_VALIDATOR_SCAN_PHOTOMETRICS, input_data)

    def valid(self) -> bool:
        _input: Image = self.input_data
        _imd: ImageMetadata = _input.metadata
        _pmetric = _imd.photometric_interpretation
        if _pmetric not in PHOTOMETRICS:
            self.invalids.append(Invalid(self.label, f"{_pmetric} != {PHOTOMETRICS}"))
        if _pmetric == 2 and _input.profile != ADOBE_PROFILE_NAME:
            self.invalids.append(Invalid(self.label, f"{_pmetric} != {ADOBE_PROFILE_NAME}"))
        return super().valid()


# determine fallback if no Validators provided
DEFAULT_VALIDATORS = [
    LABEL_VALIDATOR_SCAN_CHANNEL,
    LABEL_VALIDATOR_SCAN_COMPRESSION,
    LABEL_VALIDATOR_SCAN_RESOLUTION,
    LABEL_VALIDATOR_SCAN_PHOTOMETRICS,
]


class CombinedScanValidator(Validator):
    """Encapsulate image validation"""

    def __init__(self, path_input:Path, validator_labels:List[str]):
        super().__init__(LABEL_VALIDATOR_SCAN_COMBINED, path_input)
        self.path_input = path_input
        self.validator_labels: List[str] = validator_labels
        self.img_data: Image = None

    def valid(self) -> bool:
        self.img_data = Image(self.input_data)
        self.img_data.url = self.path_input
        self.img_data.read()
        for _label in self.validator_labels:
            _val_clazz = ValidatorFactory.get(_label)
            _val:Validator = _val_clazz(self.img_data)
            if not _val.valid():
                self.invalids.extend(_val.invalids)
        return super().valid()


# module validator mapping
VALIDATORS = {
    LABEL_VALIDATOR_SCAN_FILEDATA: ScanFileValidator,
    LABEL_VALIDATOR_SCAN_CHANNEL: ScanChannelValidator,
    LABEL_VALIDATOR_SCAN_COMPRESSION: ScanCompressionValidator,
    LABEL_VALIDATOR_SCAN_RESOLUTION: ScanResolutionValidator,
    LABEL_VALIDATOR_SCAN_FILEDATA: TiffImageFile,
    LABEL_VALIDATOR_SCAN_PHOTOMETRICS: ScanPhotometricValidator,
}


class ValidatorFactory:

    @staticmethod
    def get(validator_label:str) -> Validator:
        """Get Validator object for label"""
        
        if not validator_label in VALIDATORS:
            raise Exception(f"Missing Implementation for {validator_label}!")
        return VALIDATORS[validator_label]


def validate_tiff(tif_path, required_validators=None):
    """Ensure provided image data contains
    required metadata information concerning
    resolution and alike

    High-level validation API to trigger
    comma-separated labelled Validators
    """

    if required_validators is None:
        required_validators = DEFAULT_VALIDATORS
    _image_val = CombinedScanValidator(tif_path, required_validators)
    _image_val.valid()
    return _image_val



"""Common Validation API"""

from abc import (
    abstractmethod,
)
from dataclasses import (
    dataclass,
)
from typing import (
    List,
)


UNSET_LABEL = 'n.a.'
UNSET_NUMBR = -1
INVALID_UNSET_LABEL = 'UNSET'

LABEL_VALIDATOR_SCAN_COMBINED = 'ScanCombinedValidator'
LABEL_VALIDATOR_SCAN_CHANNEL = 'ScanMetadataChannel'
LABEL_VALIDATOR_SCAN_COMPRESSION = 'ScanMetadataCompression'
LABEL_VALIDATOR_SCAN_RESOLUTION = 'ScanMetadataResolution'
LABEL_VALIDATOR_SCAN_FILEDATA = 'ScanFiledata'
LABEL_VALIDATOR_SCAN_PHOTOMETRICS = 'ScanPhotometrics'


@dataclass
class Invalid:
    """Container for invalid data"""

    label: str
    info: str


class Validator:
    """Common Base Interface"""

    def __init__(self, label: str, input_data):
        self.label = label
        self.input_data = input_data
        self.invalids: List[Invalid] = []

    @abstractmethod
    def valid(self) -> bool:
        """Specific implementation is subject 
        of concrete Validator Implementation
        """

        return len(self.invalids) == 0



from .common import (
	LABEL_VALIDATOR_SCAN_CHANNEL,
	LABEL_VALIDATOR_SCAN_COMPRESSION,
    LABEL_VALIDATOR_SCAN_RESOLUTION,
	LABEL_VALIDATOR_SCAN_FILEDATA,
    UNSET_LABEL,
    UNSET_NUMBR,
	INVALID_UNSET_LABEL,
    Validator,
	Invalid,
)

from .fsdata import (
    FSReadException,
    FSWriteException,
    resource_can_be,
    group_can_read,
    group_can_write,
)

from .imgdata import (
    Image,
    ImageMetadata,
    InvalidImageDataException,
	CombinedScanValidator,
    ScanChannelValidator,
    ScanResolutionValidator,
    ValidatorFactory,
    validate_tiff,
)



"""API for validation on file-system-level"""

import os
import shutil
import stat

from pathlib import (
    Path,
)

from digiflow.validate import (
    group_can_read,
    group_can_write,
    resource_can_be,
)

from .conftest import (
    TEST_RES,
)


def test_can_read_common_file(tmp_path):
    """Ensure temporay test data has expected
    file attributes and can therefore be read
    """

    # arrange
    file_name = 'k2_mets_morbio_1748529021.xml'
    mets_source = Path(TEST_RES) / file_name
    mets_target = Path(str(tmp_path), file_name)
    shutil.copy(str(mets_source), str(mets_target))

    # act
    assert mets_target.exists()
    assert group_can_read(mets_target)
    # user and group may read
    assert resource_can_be(mets_target, [stat.S_IRUSR, stat.S_IRGRP])
    # user and group and even others may read
    assert resource_can_be(mets_target, [stat.S_IRUSR, stat.S_IRGRP, stat.S_IROTH])
    # but others are not allowed to write
    assert not resource_can_be(mets_target, [stat.S_IWOTH])
    assert not resource_can_be(
        mets_target, [stat.S_IRUSR, stat.S_IRGRP, stat.S_IROTH, stat.S_IWOTH])


def test_group_cant_access_resource(tmp_path):
    """Try to construct common scenario where
    resource cant be accessed by the group
    but only be the user
    """

    # arrange
    file_name = 'k2_mets_morbio_1748529021.xml'
    mets_source = Path(TEST_RES) / file_name
    mets_target = Path(str(tmp_path), file_name)
    shutil.copy(mets_source, mets_target)

    # act
    os.chmod(mets_target, stat.S_IRUSR | stat.S_IWUSR)

    # assert
    assert not group_can_read(mets_target)
    assert not group_can_write(mets_target)
    assert resource_can_be(mets_target, [stat.S_IRUSR, stat.S_IWUSR])



"""Helper to trigger DDB-Validation for digitized prints and newspaper issues/additionals
cf. https://github.com/Deutsche-Digitale-Bibliothek/ddb-metadata-schematron-validation
"""

import os

from typing import (
    Dict,
    List,
)

from pathlib import (
    Path
)

from lxml import etree as ET

from .digiflow_metadata import (
    XMLNS,
)

# trigger saxon API
SAXON_ENABLED = True
try:
    from saxonche import (
        PySaxonProcessor,
        PyXdmValue,
    )
except ModuleNotFoundError:
    SAXON_ENABLED = False

# add schematron validation language mapping
XMLNS['svrl'] = 'http://purl.oclc.org/dsdl/svrl'


# DDB Report information *not* to care about
# because we're using this for intermediate
# digitalization / migration results
DDB_IGNORE_RULES_BASIC = [
    'fileSec_02',       # fatal: no mets:fileSec[@TYPE="DEFAULT"]
    'identifier_01',    # info:  record identifier types like "bv" or "eki" not accepted
    'titleInfo_02',     # fatal: parts of work lack titel
]

# some special corner cases, when certain DDB-rules
# shall be ignored on this stage
# example:
# at migration time there's no valid href
# for the METS-pointer in the logical section, just
# the name of the SAF because Share_it creates this
# link afterwards beyond migration workflow
DDB_IGNORE_RULES_MVW = DDB_IGNORE_RULES_BASIC + [
    # fatal: metsPtr xlink:href must have valid URN
    # at import time it references local host SAF
    'structMapLogical_17',
    # error: no image from fileGroup@USE='DEFAULT' linked
    # at import time it's not yet existing
    'structMapLogical_22',
]

# periodicals (end therefore periodical_volumes)
# may have lots of extended dateIssued elements
# which seem to confuse DDB Validation
DDB_IGNORE_RULES_NEWSPAPERS = [
    'originInfo_01',
]

DDB_MEDIA_XSL = 'ddb_validierung_mets-mods-ap-digitalisierte-medien.xsl'
DDB_NEWSP_XSL = 'ddb_validierung_mets-mods-ap-digitalisierte-zeitungen.xsl'
XSL_DIR = Path(__file__).parent / 'resources' / 'xsl'
PATH_MEDIA_XSL = XSL_DIR / DDB_MEDIA_XSL
PATH_NEWSP_XSL = XSL_DIR / DDB_NEWSP_XSL

# default temporary report file
REPORT_FILE_XSLT = 'report_xslt.xml'

# which types require special treatment?
# if these show up, switch validation logics
DIGIS_MULTIVOLUME = ['Ac', 'Af', 'AF', 'Hc', 'Hf', 'HF',
                     'volume', 'periodical', 'periodical_volume']
DIGIS_NEWSPAPER = ['issue', 'additional', 'OZ', 'AZ']

# context text len
ASSERT_DSCR_MAX_LEN = 64
ASSERT_TEXT_MAX_LEN = 256
CRITICAL_VALIDATION_ROLES = ['critical', 'error', 'fatal']
FAILED_ASSERT_ERROR = 'failed_assert_error'
FAILED_ASSERT_OTHER = 'failed_assert_other'


class DigiflowTransformException(Exception):
    """Mark Missing Transformation Result"""


class DigiflowDDBException(Exception):
    """Mark Validation Errors"""


class FailedAssert:
    """Encapsulate individual Schematron Report
    Element svrl:failed-assert to gather
    @id, @role and @location Attributes
    as well as optional children svrl:text,
    svrl:description ans svrl:property
    """

    def __init__(self, elem: ET.Element):
        self.id: str = elem.get('id')
        self.role: str = elem.get('role')
        self.location: str = elem.get('location')
        self.text:str = elem.get('text')
        self.context: str = ''
        self._text: str = '. '.join(elem.xpath('svrl:text/text()', namespaces=XMLNS))
        self._description = ' '.join(elem.xpath('svrl:description/text()', namespaces=XMLNS))
        self._properties = elem.xpath('svrl:property', namespaces=XMLNS)

    def explain(self):
        """Explain the failure"""
        _msg = f'[{self.id}] '
        if len(self._properties) > 0:
            _props = '=>'.join(f"{_p.get('id')}:{_p.text}" for _p in self._properties)
            _msg = f'{_msg} {_props}'
        if len(self.context) > 0:
            _ctx = self.context
            if len(_ctx) > ASSERT_DSCR_MAX_LEN:
                _ctx = _ctx[:ASSERT_DSCR_MAX_LEN]
            _msg = f'{_msg} test:{_ctx}'
        _add_dscr = ''
        _add_text = ''
        if self._description:
            _dscr = self._description
            if len(_dscr) > ASSERT_DSCR_MAX_LEN:
                _dscr = _dscr[:ASSERT_DSCR_MAX_LEN]
            _add_dscr = _dscr + ', '
        if self._text:
            _txt = self._text
            if len(_txt) > ASSERT_TEXT_MAX_LEN:
                _txt = _txt[:(ASSERT_TEXT_MAX_LEN-3)] + '...'
            _add_text = _txt
        if len(_add_dscr) > 2 or len(_add_text) > 0:
            _msg = f'{_msg} ({_add_dscr}{_add_text})'
        return _msg

    def add_context(self, ctx: str):
        """Add even more context on error data"""
        if ctx is not None and len(ctx) > 0:
            self.context = ctx

    def is_error(self):
        """Need special care?"""
        return self.role in CRITICAL_VALIDATION_ROLES


def gather_failed_asserts(path_mets, processor, path_report, ignore_rules=None):
    """Gather all information about failed assertions"""

    _failures = _get_failures(path_mets, processor, path_report, ignore_rules)
    _aggregated = _failed_asserts_to_dict(_failures)
    if FAILED_ASSERT_ERROR in _aggregated:
        raise DigiflowDDBException(_aggregated[FAILED_ASSERT_ERROR])
    os.unlink(path_report)
    return _aggregated


def _get_failures(path_input, proc, path_report, ignores=None) -> List[FailedAssert]:
    """Inspect results from tmp report file
    if any irregularities detected, apply XSLT2.0
    expressions from report file to gather details"""

    _failures = []
    if ignores is None:
        ignores = []
    tmp_root = ET.parse(path_report).getroot()
    _failed_assert_roles = tmp_root.findall('svrl:*[@role]', XMLNS)
    if len(_failed_assert_roles) > 0:
        _fails = [FailedAssert(e)
                  for e in _failed_assert_roles
                  if e.attrib['id'] not in ignores]
        try:
            if not isinstance(path_input, str):
                path_input = str(path_input)
            _mets_doc = proc.parse_xml(xml_file_name=path_input)
            _xp_proc = proc.new_xpath_processor()
            _xp_proc.set_context(xdm_item=_mets_doc)
            for _fail in _fails:
                if _fail.location is not None:
                    _info = _xp_proc.evaluate(_fail.location)
                    if isinstance(_info, PyXdmValue) and _info.size > 0:
                        _items = [_info.item_at(i) for i in range(0, _info.size)]
                        _ctx = ','.join(_i.string_value.strip() for _i in _items)
                        _fail.add_context(_ctx)
                _failures.append(_fail)
        except Exception as _exc:
            raise RuntimeError(_exc) from _exc
    return _failures


def _failed_asserts_to_dict(fails: List[FailedAssert]) -> Dict:
    _dict = {}
    for _f in fails:
        _xpln = _f.explain()
        _role = FAILED_ASSERT_ERROR if _f.is_error() else FAILED_ASSERT_OTHER
        if _role in _dict:
            _prev = _dict[_role]
            _prev.append(_xpln)
            _dict[_role] = _prev
        else:
            _dict.setdefault(_role, [_xpln])
    return _dict


def ddb_validation(path_mets, digi_type='Aa', ignore_rules=None, post_process=gather_failed_asserts):
    """Process METS/MODS Validation which complies to Deutsche Digitale Bibliothek (DDB)
    plus applying additional set of rules which shall not be taken into account

    Further information can be found at:
        https://github.com/Deutsche-Digitale-Bibliothek/ddb-metadata-schematron-validation

    Args:
        path_mets (str|Path): resource for METS-file
        digi_type (str, optional): PICA-Type. Defaults to 'Aa'.
        ignore_rules (_type_, optional): Which Rules shall be
            ignored when evaluating validation outcome.
            Defaults to DDB_IGNORE_RULES_BASIC.
    """
    if ignore_rules is None:
        ignore_rules = DDB_IGNORE_RULES_BASIC
    if digi_type in DIGIS_MULTIVOLUME:
        ignore_rules = DDB_IGNORE_RULES_MVW
    _path_xslt = str(PATH_MEDIA_XSL)
    if digi_type in DIGIS_NEWSPAPER:
        _path_xslt = str(PATH_NEWSP_XSL)
    if not isinstance(path_mets, str):
        path_mets = str(path_mets)
    # just create saxon processor context once
    # several calls seem to break stuff
    with PySaxonProcessor() as proc:
        _path_result_file = _transform_to(path_mets, proc, _path_xslt)
        if not os.path.isfile(_path_result_file):
            raise DigiflowDDBException(f"missing tmp output file {_path_result_file}")
        return post_process(path_mets, proc, _path_result_file, ignore_rules)


def apply(path_input_xml, path_xslt, path_result_file=None, post_process=None) -> Path:
    """Low-level API to apply XSLT-Transformation
    like Schematron Validation Report Language XSLT
    on given input file and store outcome in local
    result file"""

    with PySaxonProcessor() as proc:
        _path_result_file = _transform_to(
            path_input_xml, proc, path_xslt, path_result_file)
        if not os.path.isfile(_path_result_file):
            raise DigiflowTransformException(f"missing result {_path_result_file}")
        if post_process is not None:
            return post_process(path_input_xml, proc, _path_result_file)
        return _path_result_file


def _transform_to(path_input_file, proc, path_template, path_result_file=None):
    if not isinstance(path_input_file, str):
        path_input_file = str(path_input_file)
    if not isinstance(path_template, str):
        path_template = str(path_template)
    if path_result_file is None:
        _the_dir = os.path.dirname(path_input_file)
        path_result_file = os.path.join(_the_dir, REPORT_FILE_XSLT)
    if not isinstance(path_result_file, str):
        path_result_file = str(path_result_file)
    try:
        xsltproc = proc.new_xslt30_processor()
        _exec = xsltproc.compile_stylesheet(stylesheet_file=path_template)
        _exec.transform_to_file(source_file=path_input_file,
                                output_file=path_result_file)
    except Exception as _exc:
        raise DigiflowTransformException(_exc) from _exc
    return Path(path_result_file).resolve()
